#pragma once

namespace pco
{
  constexpr uint8_t  UNDEF_B = UCHAR_MAX;
  constexpr uint16_t UNDEF_W = USHRT_MAX;
  constexpr uint32_t UNDEF_DW = UINT_MAX;
  constexpr size_t   UNDEF_QW = SIZE_MAX;

  using image16_t = std::vector<WORD>;
  using image8_t = std::vector<BYTE>;

  struct ConvertControlMono;
  struct ConvertControlColor;
  struct ConvertControlPseudoColor;
  using ConvertControl = std::variant<ConvertControlMono, ConvertControlColor, ConvertControlPseudoColor>;

  // helper type for the visitor #4
  template<class... Ts> struct overloaded : Ts... { using Ts::operator()...; };
  // explicit deduction guide (not needed as of C++20)
  template<class... Ts> overloaded(Ts...)->overloaded<Ts...>;

  enum class DataFormat {
    Undefined,
    Mono8, // 8 bit camera, compressed images
    Mono16,
    BGR8,
    BGRA8,
    BGR16,
    CompressedMono8 // 
  };

  enum class RawFormat {
    Word,
    Byte
  };

  static constexpr size_t WAIT_TIME_MS = 1;  //1ms delay, for wait funcs;

  enum class RecordMode {
    sequence, sequence_non_blocking, ring_buffer, fifo,
    sequence_dpcore, sequence_non_blocking_dpcore, ring_buffer_dpcore, fifo_dpcore,
    tif, multitif, pcoraw, b16, dicom, multidicom,
    camram_ring, camram_segment
  };

  enum class CameraInterface : WORD {
    FireWire            = PCO_INTERFACE_FW     ,   // Firewire interface
    CameraLinkMTX       = PCO_INTERFACE_CL_MTX ,   // Cameralink Matrox Solios / Helios
    GenICam             = PCO_INTERFACE_GENICAM,   // Genicam wrapper interface
    CameraLinkNAT       = PCO_INTERFACE_CL_NAT ,   // Cameralink National Instruments
    GigE                = PCO_INTERFACE_GIGE   ,   // Gigabit Ethernet
    USB                 = PCO_INTERFACE_USB    ,   // USB 2.0
    CameraLinkME4       = PCO_INTERFACE_CL_ME4 ,   // Cameralink Silicon Software Me4
    USB3                = PCO_INTERFACE_USB3   ,   // USB 3.0 and USB 3.1 Gen1
    WLAN                = PCO_INTERFACE_WLAN   ,   // WLan (Only control path, not data path)
    CLHS                = PCO_INTERFACE_CLHS   ,   // Cameralink HS
    Any                 = UNDEF_W
  };

  struct Binning
  {
    WORD vert = UNDEF_W;
    WORD horz = UNDEF_W;
  };

  struct Roi	//region of interest
  {
    size_t x0 = UNDEF_QW;
    size_t y0 = UNDEF_QW;
    size_t x1 = UNDEF_QW;
    size_t y1 = UNDEF_QW;
    size_t width() { return x1 - x0 + 1; }
    size_t height() { return y1 - y0 + 1; }
    size_t size() { return width() * height(); }
    size_t evenPaddedWidth() { return width() + (4 - width() % 4); }
    size_t paddedSize() { return evenPaddedWidth() * height(); }
  };

  struct Configuration
  {
    double exposure_time_s = NAN;
    double delay_time_s = NAN;
    Roi roi;
    WORD timestamp_mode = UNDEF_W;
    DWORD pixelrate = UNDEF_DW;
    WORD trigger_mode = UNDEF_W;
    WORD acquire_mode = UNDEF_W;
    WORD metadata_mode = UNDEF_W;
    WORD noise_filter_mode = UNDEF_W;
    Binning binning;
  };

  struct Description
  {
    // properties
    DWORD serial = UNDEF_DW;
    WORD type = 0;
    WORD sub_type = 0;
    CameraInterface interface_type = CameraInterface::Any;
    // exposure_time_s
    double min_exposure_time_s = NAN;
    double max_exposure_time_s = NAN;
    double min_exposure_step_s = NAN;
    // delay_time_s
    double min_delay_time_s = NAN;
    double max_delay_time_s = NAN;
    double min_delay_step_s = NAN;
    // roi
    size_t min_width = UNDEF_QW;
    size_t min_height = UNDEF_QW;
    size_t max_width = UNDEF_QW;
    size_t max_height = UNDEF_QW;
    size_t roi_step_horz = UNDEF_QW;
    size_t roi_step_vert = UNDEF_QW;
    bool roi_symmetric_horz = false;
    bool roi_symmetric_vert = false;
    // bitness
    WORD bit_resolution = UNDEF_W;
    //timestamp_mode
    bool has_timestamp_mode = false;
    bool has_timestamp_mode_ascii_only = false;
    // pixelrate
    std::vector<DWORD> pixelrate_vec;
    // trigger_mode -> has no descriptor dependency
    // acquire_mode
    bool has_acquire_mode = false;
    bool has_ext_acquire_mode = false;
    // metadata
    bool has_metadata_mode = false;
    bool has_ram = false;
    // binning
    std::vector<WORD> binning_horz_vec;
    std::vector<WORD> binning_vert_vec;
  };


  struct ConvertControlMono
  {
    bool sharpen = true;
    bool adaptive_sharpen = true; 
    bool flip_vertical = true;
    bool auto_minmax = true;
    int min_limit = 0;
    int max_limit = 20000;
    double gamma = 1.0;
    int contrast = 0;
  };
  struct ConvertControlColor
  {
    bool sharpen = true;
    bool adaptive_sharpen = true; 
    bool flip_vertical = true;
    bool auto_minmax = true;
    int min_limit = 0;
    int max_limit = 20000;
    double gamma = 1.0;
    int contrast = 0;
    bool pco_debayer_algorithm = false;
    int color_temperature = 6500;
    int color_saturation = 0;
    int color_vibrance = 0;
    int color_tint = 0;
  };
  struct ConvertControlPseudoColor
  {
    bool sharpen = true;
    bool adaptive_sharpen = true; 
    bool flip_vertical = true;
    bool auto_minmax = true;
    int min_limit = 0;
    int max_limit = 20000;
    double gamma = 1.0;
    int contrast = 0;
    int color_temperature = 6500;
    int color_saturation = 0;
    int color_vibrance = 0;
    int color_tint = 0;
    std::filesystem::path lut_file;
  };

  // free functions

  inline int bcd_counter(BYTE* bcds)
  {
    auto bcd_to_decimal = [](BYTE bcd) {
      return int(bcd / 0x10) * 10 + (bcd % 0x10);
    };

    int ret = bcd_to_decimal(bcds[0])
      + 100 * bcd_to_decimal(bcds[1])
      + 10'000 * bcd_to_decimal(bcds[2])
      + 1000'000 * bcd_to_decimal(bcds[3]);

    return ret;
  }

};
