#include "stdafx.h"

#include "defs.h"
#include "cameraexception.h"

#include "image.h"
#include "camera.h"


namespace pco {

  Image::Image()
  {}

  Image::Image(Roi roi, DataFormat data_format, RawFormat raw_format)
    : roi_(roi), data_format_(data_format)
  {
    if (raw_format == RawFormat::Word)
      raw_data_ = image16_t(roi.size());
    if (raw_format == RawFormat::Byte)
      raw_data_ = image8_t(roi.size());

    switch (data_format_)
    {
    case pco::DataFormat::Undefined:
      break; // data still monostate
    case pco::DataFormat::Mono8:
      data_ = image8_t(roi.paddedSize());
      break;
    case pco::DataFormat::Mono16:
      break; // data not needed
    case pco::DataFormat::BGR8:
      data_ = image8_t(3 * roi.paddedSize());
      break;
    case pco::DataFormat::BGRA8:
      data_ = image8_t(4 * roi.size());
      break;
    case pco::DataFormat::BGR16:
      data_ = image16_t(3 * roi.paddedSize());
      break;
    case pco::DataFormat::CompressedMono8:
      data_ = image8_t(roi.size());
      break;
    default:
      break; // data still monostate
    }
  }

  Image::~Image()
  {
  }

  bool Image::is8Bit()
  {
    std::vector<DataFormat> data_8bit{
      DataFormat::Mono8,
      DataFormat::BGR8,
      DataFormat::BGRA8,
      DataFormat::CompressedMono8
    };
    if (std::find(data_8bit.begin(), data_8bit.end(), data_format_) != data_8bit.end())
      return true;
    return false;
  }

  bool Image::is16Bit()
  {
    std::vector<DataFormat> data_8bit{
      DataFormat::Mono16,
      DataFormat::BGR16
    };
    if (std::find(data_8bit.begin(), data_8bit.end(), data_format_) != data_8bit.end())
      return true;
    return false;
  }

  bool Image::isColored()
  {
    if (data_format_ == DataFormat::Undefined || data_format_ == DataFormat::Mono16
      || data_format_ == DataFormat::Mono8 || data_format_ == DataFormat::CompressedMono8)
    {
      return false;
    }
    return true;
  }

  DataFormat Image::getDataFormat()
  {
    return data_format_;
  }

  size_t Image::width()
  {
    return roi_.width();
  }

  size_t Image::height()
  {
    return roi_.height();
  }

  bool Image::validAllocation(Roi roi, DataFormat data_format, RawFormat raw_format)
  {
    if (data_format != data_format_)
      return false;

    // raw data
    if (std::holds_alternative<std::monostate>(raw_data_))
      return false;
    if (raw_format == RawFormat::Word)
      if (!std::holds_alternative<image16_t>(raw_data_))
        return false;
    if (raw_format == RawFormat::Byte)
      if (!std::holds_alternative<image8_t>(raw_data_))
        return false;

    if (size() != roi.size()) // checks size of raw_data_
      return false;

    // data
    switch (data_format_)
    {
    case pco::DataFormat::Undefined:
      break; // data still monostate
    case pco::DataFormat::Mono8:
      if (std::holds_alternative<image8_t>(data_))
        if (std::get<image8_t>(data_).size() == roi.paddedSize())
          return true;
      break;
    case pco::DataFormat::Mono16:
      if (size() == roi.size()) // checks raw_data_
        return true;
      break; // data_ not needed
    case pco::DataFormat::BGR8:
      if (std::holds_alternative<image8_t>(data_))
        if (std::get<image8_t>(data_).size() == 3 * roi.paddedSize())
          return true;
      break;
    case pco::DataFormat::BGRA8:
      if (std::holds_alternative<image8_t>(data_))
        if (std::get<image8_t>(data_).size() == 4 * roi.size())
          return true;
      break;
    case pco::DataFormat::BGR16:
      if (std::holds_alternative<image16_t>(data_))
        if (std::get<image16_t>(data_).size() == 3 * roi.paddedSize())
          return true;
      break;
    case pco::DataFormat::CompressedMono8:
      if (std::holds_alternative<image8_t>(data_))
        if (std::get<image8_t>(data_).size() == roi.size())
          return true;
      break;
    default:
      break; // data still monostate
    }

    return false;
  }

  void Image::resize(Roi roi, DataFormat data_format, RawFormat raw_format)
  {
    switch (data_format)
    {
    case pco::DataFormat::Undefined:
      return;
      break; // data still monostate
    case pco::DataFormat::Mono8:
      if (std::holds_alternative<image8_t>(data_))
        std::get<image8_t>(data_).resize(roi.paddedSize());
      else
        data_ = image8_t(roi.paddedSize());
      break;
    case pco::DataFormat::Mono16:
      break; // data not needed
    case pco::DataFormat::BGR8:
      if (std::holds_alternative<image8_t>(data_))
        std::get<image8_t>(data_).resize(3 * roi.paddedSize());
      else
        data_ = image8_t(3 * roi.paddedSize());
      break;
    case pco::DataFormat::BGRA8:
      if (std::holds_alternative<image8_t>(data_))
        std::get<image8_t>(data_).resize(4 * roi.size());
      else
        data_ = image8_t(4 * roi.size());
      break;
    case pco::DataFormat::BGR16:
      if (std::holds_alternative<image16_t>(data_))
        std::get<image16_t>(data_).resize(3 * roi.paddedSize());
      else
        data_ = image16_t(3 * roi.paddedSize());
      break;
    case pco::DataFormat::CompressedMono8:
      if (std::holds_alternative<image8_t>(data_))
        std::get<image8_t>(data_).resize(roi.size());
      else
        data_ = image8_t(roi.size());
      break;
    default:
      return;
      break; // data still monostate
    }

    if (raw_format == RawFormat::Word)
    {
      if (std::holds_alternative<image16_t>(raw_data_))
        std::get<image16_t>(raw_data_).resize(roi.size());
      else
        raw_data_ = image16_t(roi.size());
    }
    else if (raw_format == RawFormat::Byte)
    {
      if (std::holds_alternative<image8_t>(raw_data_))
        std::get<image8_t>(raw_data_).resize(roi.size());
      else
        raw_data_ = image8_t(roi.size());
    }

    data_format_ = data_format;
    roi_ = roi;
  }

  void Image::setRecorderImageNumber(DWORD recorder_image_number)
  {
    recorder_image_number_ = recorder_image_number;
  }

  DWORD Image::getRecorderImageNumber()
  {
    return recorder_image_number_;
  }

  void Image::setMetaData(PCO_METADATA_STRUCT _meta_data)
  {
    meta_data = _meta_data;
  }

  void Image::clearMetaData()
  {
    meta_data = std::nullopt;
  }

  PCO_METADATA_STRUCT* Image::getMetaDataPtr()
  {
    if (meta_data.has_value() && meta_data.value().wSize)
      return &meta_data.value();
    else
      return nullptr;
  }

  PCO_METADATA_STRUCT Image::getMetaData()
  {
    if (!meta_data.has_value())
      throw CameraException("pco::Image does not hold meta data");
    return meta_data.value();
  }

  void Image::setTimestamp(PCO_TIMESTAMP_STRUCT _timestamp)
  {
    if (!timestamp.has_value())
      timestamp = _timestamp;
  }

  void Image::clearTimestamp()
  {
    timestamp = std::nullopt;
  }

  PCO_TIMESTAMP_STRUCT* Image::getTimestampPtr()
  {
    if (timestamp.has_value() && timestamp.value().wSize)
      return &timestamp.value();
    else
      return nullptr;
  }

  std::pair<void*, size_t> Image::raw_data()
  {
    void* d = nullptr;
    std::visit(overloaded{
      [](std::monostate) {},
      [&d](image8_t& v) {d = static_cast<void*>(v.data()); },
      [&d](image16_t& v) {d = static_cast<void*>(v.data()); }
      }, raw_data_);

    size_t s = 0;
    std::visit(overloaded{
      [](std::monostate) {},
      [&s](image8_t& v) {s = v.size(); },
      [&s](image16_t& v) {s = v.size() * sizeof(WORD); }
      }, raw_data_);
    return { d, s };
  }

  std::pair<void*, size_t> Image::data()
  {
    void* d = nullptr;
    size_t s = 0;
    if (data_format_ == DataFormat::Mono16)
    {
      std::visit(overloaded{
        [](std::monostate) {},
        [&d](image8_t& v) {d = v.data(); },
        [&d](image16_t& v) {d = v.data(); }
        }, raw_data_);

      std::visit(overloaded{
        [](std::monostate) {},
        [&s](image8_t& v) {s = v.size(); },
        [&s](image16_t& v) {s = v.size() * sizeof(WORD); }
        }, raw_data_);
    }
    else
    {
      std::visit(overloaded{
        [](std::monostate) {},
        [&d](image8_t& v) {d = v.data(); },
        [&d](image16_t& v) {d = v.data(); }
        }, data_);

      std::visit(overloaded{
        [](std::monostate) {},
        [&s](image8_t& v) {s = v.size(); },
        [&s](image16_t& v) {s = v.size() * sizeof(WORD); }
        }, data_);

    }
    return { d, s };
  }

  size_t Image::size()
  {
    size_t s = 0;
    std::visit(overloaded{
      [](std::monostate) {},
      [&s](image8_t& v) {s = v.size(); },
      [&s](image16_t& v) {s = v.size(); }
      }, raw_data_);
    return s;
  }

  image8_t& Image::vector_8bit()
  {
    if (std::holds_alternative<image8_t>(data_))
      return std::get<image8_t>(data_);
    throw CameraException("Image data does not contain 8bit data");
  }

  image16_t& Image::vector_16bit()
  {
    if (data_format_ == DataFormat::Mono16)
      return std::get<image16_t>(raw_data_);

    if (std::holds_alternative<image16_t>(data_))
      return std::get<image16_t>(data_);
    throw CameraException("Image data does not contain 16bit data");
  }

  image16_t& Image::raw_vector_16bit()
  {
    return std::get<image16_t>(raw_data_);
  }

}
